from os import makedirs
from pathlib import Path
from pickle import dump as pkl_dump
from pickle import load as pkl_load
from time import time

from appdirs import user_cache_dir
from domain_suffixes.suffixes import Suffixes
from humanize.time import precisedelta

PICKLE_FILEPATH = Path(
    user_cache_dir(
        appname='InspyLogger',
        appauthor='Inspyre-Softworks'
    )
).joinpath('tld.pkl')


class TLDCache(object):

    parent_dir = PICKLE_FILEPATH.parent

    def __init__(self):
        self.__created_ts = None
        self.__pickle_fp_exists = None
        self.__num_tlds = None
        self.__tlds = None
        self.__data = None
        self.__last_renewal_ts = None
        self.__renewal_hist = None

    def build(self):
        self.__data = {
                'created_ts': time(),
                'renewal_data': {
                        'last_ts': None,
                        'history': [],
                },
                'tlds': self.fetch_from_source()
        }

        return self.__data

    def load(self):
        """ Load the pickle file for the cache.

        Will attempt to load the pickle file into memory from the path indicated by :obj:`PICKLE_FILEPATH`

        Raises:
            :exception:`FileNotFoundError`: If the pickle file does not exist.

        """
        print('Loading pickle file')
        with open(PICKLE_FILEPATH, 'rb') as pfp:
            res = pkl_load(pfp)

        return res

    def write(self):
        """ Write the pickle file.

        Write the contents of the cache to the file-system path indicated by :obj:`PICKLE_FILEPATH`

        Raises:
            :exception:`PermissionError`: If the pickle file can't be written due to permissions writing the file to storage. 
        """

        if not self.parent_dir.exists():
            makedirs(self.parent_dir, exist_ok=True)

        with open(PICKLE_FILEPATH, 'wb') as pfp:
            pkl_dump(self.__data, pfp)

    def fetch_from_source(self):
        """ Fetch the top-level-domain list from the source

        Returns:
            :class:`list`: A list of all known top-level domain suffixes from the domain-suffixes package.
        """
        suffixes = Suffixes()
        return suffixes.get_all_tlds()

    @property
    def contents(self) -> dict:
        if self.__data is None:
            try:
                self.__data = self.load()
            except FileNotFoundError:
                self.__data = self.build()
                self.write()

        return self.__data

    @property
    def TLDs(self) -> list:
        """ Returns a list of TLDs (top-level domains) for validating against """
        return self.contents['tlds']

    @property
    def created_timestamp(self) -> float:
        """
        An alias for :attr:`created_ts`

        Returns:
            created_ts (float):
                The time (in seconds since Epoch) when the cache was created.

        """
        return self.created_ts

    @property
    def created_ts(self) -> float:
        """
        The time (in seconds since Epoch) when the cache was created
        """
        return self.contents['created_ts']

    @property
    def age_in_seconds(self) -> float:
        """
        The number of seconds that have elapsed since creation of the cache.
        """
        return time() - self.created_ts

    @property
    def age(self) -> str:
        return precisedelta(self.age_in_seconds)



"""

Contains exceptions for InspyLogger

"""
import inspect


class InSPyLoggerError(Exception):
    frame = inspect.stack()[1]
    caller_name = frame[3]

    # Let's set a general error message for the parent exception
    # letting the end-user know that the child exception is
    # of this type;
    p_message = f"InSPyLogger has encountered an error! [Caller: {caller_name}]\n"

    def __init__(self, message=None, skip_print=False):
        # If the child exception provides a message we can introduce this with
        # the following phrase;
        msg_prefix = f'\nSome further context...\n'

        # If we're provided a message on instantiation we'll prepend our prefix
        # to it while assigning it to the :var:`message` variable. If no message
        # was provided :var:`message' will evaluate to str('')
        message = f'{msg_prefix} - {message}' if message is not None else ''

        # Assign our prefix string to a class attribute for easy debugging later.
        self.msg_prefix = msg_prefix

        # finally; concatenate the entirety of the error message delivered on raise.
        self.message = f"{self.p_message}{self.msg_prefix}- {self.message}"

        if not skip_print:
            print(self.message)


class DeviceNotStartedError(InSPyLoggerError):
    # Let's set a more specific message for the child exception.
    message = "[Device Not Started]\n" \
              "The root logging device has not yet been started! Try `isl.device.start()`\n    "
    prefix = "Some further information from the caller:\n"

    __child_frame = inspect.stack()[1]
    __child_frame_name = __child_frame[3]

    def __init__(self, message=None, skip_print=False):

        # Prettify our base message string.
        self.message = f"[{self.__class__.__name__}] [{self.__child_frame_name}] - {self.message}"

        message = f"{self.prefix} - {message}" if message is not None else ''

        self.message += message

        super(DeviceNotStartedError, self).__init__(message = self.message, skip_print = skip_print)


class DeviceAlreadyStartedError(InSPyLoggerError):
    # Let's set a more specific message for the child exception.
    message = "[Device Already Started] The device is already started.\n" \
              "The logging device is already instantiated and started."
    prefix = "Some futher information from the caller:\n"

    def __init__(self, message=None, skip_print=False):
        self.__child_caller_frame = inspect.stack()[1]
        self.__child_caller_name = self.__child_caller_frame[3]

        # Prettify our base message string.
        self.message = f"[{self.__class__.__name__}] - {self.message}"

        message = f"{self.prefix} - {message}" if message is not None else ''

        self.message += message

        super(DeviceAlreadyStartedError, self).__init__(message = self.message, skip_print = skip_print)



class ManifestEntryExistsError(Exception):
    def __init__(self, msg=None, caller_name=None):
        """
        
        Raise an exception advising an attempted log creation caller that there's already an entry for this logger on the manifest.

        """

        if msg is None:
            msg = "You've attempted to add a a previously existing logger to the log manifest"

        if caller_name is None:
            frame = inspect.stack()[1]
            caller_name = frame[3]

        self.message = str(f"{msg} | Caller: {caller_name}")



#!/usr/bin/env python3
"""

This package contains a class that will create and maintain a logging device for you

Returns:
    InspyLogger: A colored and formatted logging device.

"""
from __future__ import annotations
import inspect
import logging
import sys
import time
from logging import DEBUG, INFO, Logger, WARNING, getLogger
from urllib.error import URLError

from colorlog import ColoredFormatter
from luddite import get_version_pypi
from pkg_resources import DistributionNotFound

from inspy_logger.errors import ManifestEntryExistsError, DeviceNotStartedError, DeviceAlreadyStartedError
from inspy_logger.manifest import Manifest
from inspy_logger.helpers.network import check_connectivity

from inspy_logger.__about__ import __version__ as VERSION, __prog__ as PROG, __authors__ as AUTHORS

#####################################
## MOST ACCURATE VERSION INDICATOR ##
#####################################

RELEASE = "2.1"

ON_REPO = True

#####################################
######## VERSION DATA ABOVE! ########
#####################################

pretty_name = "InSPy Logger"
PKG_NAME = __package__
PY_VER = sys.version.split(" ")[0]

LEVELS = ["debug", "info", "warning"]
"""The names of the log output levels one can pick from"""

latest_version = "Please start the InspyLogger class"

islatest = None


class InspyLogger(object):
    """ InspyLogger is meant to be a shortcut to a program-wide logger with levels and ease-of-use. """

    class Version:

        def __init__(self):
            """

            This class helps manage the Version information for InspyLogger

            """

            self.package_name = "InSPy-Logger"
            """The name of the program"""

            self.local = VERSION
            """
            The hard-coded version of the program. Unless this is changed by someone 
            other than the developers 
            """

            self.__is_up_to_date = None

            self.__needs_update = None
            self.__offline = None

            self.optional_update = None
            self.latest_stable = None
            self.latest_pr = None
            self.offline = False

        @property
        def needs_update(self):
            """
            The needs_update function checks whether the package version is up to date with
            the latest version on PyPI. If it is not, then it returns True, otherwise False.

            Returns:
                * True:
                    The local package version does not match the latest version available via Pypi.
                * False:
                    The local package version matches or exceeds the latest version available via Pypi.
            """
            if self.__needs_update is None or self.__is_up_to_date is None:
                return self.is_up_to_date

        @staticmethod
        def __instruction_feeder(instruct_group):
            for line in instruct_group:
                print(line)

        @property
        def is_up_to_date(self):
            """

            Checks with PyPi to make sure you have the latest version.

            Returns:

                * This function will return three fields with varying values which are the following in respective order:
                    - Is the local version at least up to date with the latest stable version found on PyPi? (In the form of a boolean)
                    - A string that best matches the update status of InspyLogger.
                        * The possible values are:

                            - match: Matches the latest stable version, not pre-release

                            - pr_ver: The version number is a pre-release, as it's number is greater than the latest stable copy available on Pypi

                            - not_released: The version number not only exceeds the latest stable version on PyPi but also the latest Pre-Release copy available on PyPi

                            - outdated: The version number is less than the latest stable version available on PyPi

                One of three possible results:


                    - (bool : True, str : `match`, obj : VERSION)

                        Indicates that this version number matches the copy on PyPi

                    - (True (bool), "pr_ver", VERSION):
                    Indicates that a "Pre-Release" version of InSPy-Logger is being used and matches a release found on PyPi
                    - (False (bool), "not_released", VERSION): Indicates that this version number surpasses the highest available on PyPi
                    - (False (bool), "outdated", VERSION):  Indicates that this version number is lower than the latest stable version on PyPi

            """

            latest = self.get_latest()

            if latest is None and self.offline:
                self.latest_stable = ("Unknown", "Offline")
                self.latest_pr = ("Unknown", "Offline")
                self.__needs_update = ("Unknown", "Offline")
                self.optional_update = False
            if not self.offline and self.needs_update is not None and self.needs_update:
                notif_lines = [
                        "An update for InSPy-Logger is available! Please update!",
                        "You can update in one of the following ways:"
                ]

                direct_update_instructions = [
                        "    - Through the package itself:",
                        "        from inspy_logger import InspyLogger, LogDevice",
                        "        ",
                        "        i_log = InspyLogger()",
                        "        iLog_ver = i_log.Version()",
                        "        iLog_ver.update(pr=False)",
                ]

                pypi_update_instructions = [
                        "    - Through your system's implementation of PIP",
                        "        $> python3 -m pip install --update inspy_logger",
                        "        OR",
                        "        $> python3 -m pip install inspy_logger VER"
                ]

                border = str("*-*" * 20)
                # Notify the user through the console of an update being available.
                print(notif_lines)

                print(border)

                self.__instruction_feeder(direct_update_instructions)

                print("-|-" * 20)

                self.__instruction_feeder(pypi_update_instructions)

                print(border)

                return False, "outdated", self.local

            # If the constant ON_REPO is Bool(False) we'll assume it's a pre-release
            if not ON_REPO:
                return True, "pr_ver", self.local

        def get_latest(self):
            try:
                return get_version_pypi(self.package_name)
            except URLError:
                statement = "Unable to access distribution server. Seeing if network is down..."
                print(statement)
                if not check_connectivity():
                    statement = "Unable to connect to the internet, therefore I can't get remote version data"
                    print(statement)
                    self.offline = False
                else:
                    statement = "Able to access the internet, however there's no access to the version data server...."
                    print(statement)
                    self.offline = True
                ver = "Unknown"
                self.__is_up_to_date = False
                self.offline = True
                return None
            except DistributionNotFound:
                ver = "Unknown"
                is_latest = False

            return ver, is_latest

    def __init__(self, log_name=None, log_level=None, autostart=False, start_on_device_call=False):
        self.__auto_start = autostart

        self.loc_version = self.Version()  # 'loc' = short for 'local' for the curious

        self.root_name = 'InspyLogger'

        self.__loc_version = None

        self.__name = log_name

        self.__level = log_level

        if log_level is None:
            self.__level = 'INFO'

        self.__device = None

        self.__VERSION = None

        self.__root_device = None

        self.__start_on_device_call = start_on_device_call

        if log_name:
            self.device = self.LogDevice(log_name, log_level)

            if self.__auto_start and not self.device.started:
                self.device.start()
        else:
            self.dev

    @property
    def name(self) -> (str | None):
        """
        The name of the logging device.
        """
        return self.__name

    @name.setter
    def name(self, new: (str | None)):
        if self.__device is not None and self.device.started:
            raise DeviceAlreadyStartedError('You cannot change the name of a running log device. Try adding a '
                                            'child device?')
        self.__name = new

    @property
    def level(self):
        return self.__level

    @level.setter
    def level(self, new):
        if self.device is None:
            raise DeviceNotStartedError()
        if isinstance(self.device, self.LogDevice):
            self.device.adjust_level(new)
        else:
            raise TypeError("device must be a LogDevice instance!")

    @property
    def autostart(self):
        return self.__auto_start

    @autostart.setter
    def autostart(self, new):
        if self.device is not None and not self.device.started:
            if isinstance(new, bool):\
                self.__auto_start = new
            else:
                raise TypeError('Autostart must be a boolean value')
        elif self.device.started:
             raise DeviceAlreadyStartedError('Setting "autostart" would do no good.')

    @property
    def start_on_device_call(self):
        return self.__start_on_device_call

    @start_on_device_call.setter
    def start_on_device_call(self, new):
        if not isinstance(new, bool):
            raise TypeError('start_on_device_call must be a boolean value!')

        self.__start_on_device_call = new

    @property
    def device(self):
        if self.__device is not None and not self.__device.started and self.start_on_device_call:
            self.device.start()

        return self.__device

    @device.setter
    def device(self, new):
        if not isinstance(new, self.LogDevice):
            raise TypeError('Device must be an InspyLogger.LogDevice instance!')

        self.__device = new

    @property
    def local_version(self):
        """
        The local_version function is a property that returns the local version of the package.
        It is used to determine if there are any updates available on PyPI.
        """

        if self.__loc_version is None:
            self.__loc_version = self.Version()

        return self.__loc_version

    @property
    def VERSION(self):
        """
        The VERSION function is used to return the version of the package.
        It will first check if a local version has been set, and if not it will
        return the pypi version.

        Returns:

            The version of the local package
        """

        if self.__VERSION is None:
            self.__VERSION = self.local_version.local

        return self.__VERSION

    # def __get_version(self, pkg_name):
    #     """
    #     The :func:`__get_version` function is used to print the version of the package and
    #     the versions available from PyPi.

    #     It also prints whether or not there are developmental versions available. If there
    #     are developmental versions, it will also print a warning that these should not be
    #     used in production.

    #     Arguments:

    #         pkg_name (str):
    #             The name of the package you're seeking the version of. (Required)

    #     Returns:
    #         A string that contains the name of the package, it's version number,
    #         and a statement about whether or not the target package's local
    #         version should be updated.

    #     """

    #     if self.loc_version.is_up_to_date:
    #         update_statement = "You are up to date!"
    #     else:
    #         if pkg_ver.parse(str(self.VERSION)) < pkg_ver.parse(str(get_version_pypi(pkg_name))):
    #             update_statement = f"You are running an older version of {pkg_name} than what is available. Consider upgrading."
    #         else:
    #             if self.VERSION in get_versions_pypi(pkg_name):
    #                 avail_ver = (
    #                     ", a developmental version available via the PyPi repository"
    #                 )
    #             else:
    #                 avail_ver = (
    #                     ", a version that is NOT available via any online package manager"
    #                 )
    #             update_statement = f"You are running a version of {pkg_name} that is newer than the latest version{avail_ver}"
    #             update_statement += f"\nThe versions available from PyPi are: {', '.join(get_versions_pypi(pkg_name))}"

    #     ver = str(f"{pretty_name} ({self.VERSION}) using Python {PY_VER}\n" + f"{update_statement}")

    #     print(ver)

    #     return ver

    class LogDevice(Logger):
        """
        Starts a colored and formatted logging device for you.

        Starts a colored and formatted logging device for you. No need to worry about handlers, etc

        Args:

            device_name (str): A string containing the name you'd like to choose for the root logger

            log_level (str): A string containing the name of the level you'd like InspyLogger to be limited to. You can choose between:

            - debug
            - info
            - warning

        """

        def __add_child(self, name: str):
            # Start a logger for this function
            log = getLogger(f"{self.own_logger_root_name}.add_child")
            ts = time.time()
            frame = inspect.stack()[1]

            frame_name = frame[3]

            line_no = frame[2]

            file_name = frame[1]

            root_name = self.root_name

            log.debug(
                "Received request to add %(name)s to %(root_name)s by %(frame_name)s on line %(line_no)s of %(file_name)s.")
            log.debug("Full Frame Info ")

            name = f'{root_name}.{name}'

            if existing := self.manifest.check(name):
                try:
                    raise ManifestEntryExistsError()
                except ManifestEntryExistsError as e:
                    log.warning(e.message)
                    return existing




            logger = getLogger(f'{name}')

            self.manifest.add(name=name, logger_device=logger, calling_file=file_name, line_num=line_no)

            return logger

        def add_child(self, name: str):
            """

            Create (and return) a child logger under the root log device. Also add it to the manifest to keep track of it.

            Args:
                name (str): The name you'd like to give the child logger

            Returns:
                getLogger: A child logging device.

            """

            return self.__add_child(name)

        def adjust_level(self, l_lvl="info", silence_notif=False):
            """

            Adjust the level of the logger associated with this instance.

            Args:
                l_lvl (str): A string containing the name of the level you'd like InspyLogger to be limited to. You can choose between:

                - debug
                - info
                - warning

                silence_notif (bool): Silence notifications (of 'info' level) when adjusting the logger's level. True for
                no output and False to get these notifications.

            Returns:
                None

            """

            _log = getLogger(self.root_name)

            _caller = inspect.stack()[1][3]

            if self.last_lvl_change_by is None:
                _log.info("Setting logger level for first time")
                _log.debug("Signing in")
                self.last_lvl_change_by = "Starting Logger"
                last_level = self.level
                last_level_change_by = self.last_lvl_change_by
            else:
                if not silence_notif:
                    _log.info(
                        "%(_caller)s is changing logger level from %(last_level)s to %(l_lvl)s",
                    )

                    _log.info(
                        "Last level change was implemented by: %(last_level_change_by)s"
                    )

                    _log.info("Updating last level changer")

                self.last_lvl_change_by = _caller

            self.level = l_lvl

            if isinstance(self.level, int):
                new_lvl = self.level
            else:
                new_lvl = None

                if self.level.lower() == "debug":
                    new_lvl = DEBUG
                elif self.level.lower() == "info":
                    new_lvl = INFO
                elif self.level.lower() in ["warn", "warning"]:
                    new_lvl = WARNING

            _log.setLevel(new_lvl)

        def start(self, mute=False, no_version=False):
            """

            Start the actual logging instance and fill the attributes that __init__ creates.

            Arguments:

                mute (bool): Mute all output that starting the root-logger would produce. True: No output on executing start() | False: Do not suppress all output

                no_version (bool): If you start the logger using the 'debug' log-level the logger will output its own version information. True: Suppress this output, no matter the log-level | False: Do no suppress this output

            Note:
                If you give the 'mute' parameter a value of `True` then the value of the `no_version` parameter will be ignored.

            Returns:
                None

            """
            if self.started:
                self.device.warning(
                    "There already is a base logger for this program. I am using it to deliver this message."
                )
                return None

            formatter = ColoredFormatter(
                "%(bold_cyan)s%(asctime)-s%(reset)s%(log_color)s::%(name)s.%(module)-14s::%(levelname)-10s%(reset)s%("
                "blue)s%(message)-s",
                datefmt = None,
                reset = True,
                log_colors = {
                        "DEBUG":    "bold_cyan",
                        "INFO":     "bold_green",
                        "WARNING":  "bold_yellow",
                        "ERROR":    "bold_red",
                        "CRITICAL": "bold_red",
                },
            )

            self.device = logging.getLogger(self.root_name)
            self.main_handler = logging.StreamHandler()
            self.main_handler.setFormatter(formatter)
            self.device.addHandler(self.main_handler)
            self.adjust_level(self.level)
            _log_ = logging.getLogger(self.own_logger_root_name)
            if not mute:
                _log_.info(f"Logger started for {self.root_name}")
                if not no_version:
                    _log_.debug(
                        f"\nLogger Info:\n" +
                        ("*" * 35) + f"\n{VERSION}\n" + ("*" * 35)
                    )
                    self.__started = True

            return self.device

        def __init__(self, log_name, log_level=None):
            """

            Starts a colored and formatted logging device for you. No need to worry about handlers, etc

            Args:

                device_name (str): A string containing the name you'd like to choose for the root logger

                log_level (str): A string containing the name of the level you'd like InspyLogger to be limited to.

                You can choose between:
                - debug
                - info
                - warning
            """
            frame = inspect.stack()[1]

            frame_name = frame[3]

            line_no = frame[2]

            file_name = frame[1]

            if log_level is None:
                log_level = 'INFO'

            self.__device_name = log_name

            self.__started = False
            self.__level = log_level

            self.root_name = self.name
            self.own_logger_root_name = f"{self.root_name}.InSPyLogger"
            super(InspyLogger.LogDevice, self).__init__(name = self.own_logger_root_name, level = self.level.upper())

            self.last_lvl_change_by = None
            self.device = None
            self.__manifest = Manifest(self.root_name, self, frame_name, line_no)
            self.main_handler = None

        @property
        def manifest(self):
            return self.__manifest

        @property
        def name(self):
            """
            The name of the logging device.

            Returns:
                (str | None): The name of the logging device.

            """
            return self.__device_name

        @name.setter
        def name(self, new):
            if not self.__device_name:
                self.__device_name = new

        @property
        def level(self):
            return self.__level

        @level.setter
        def level(self, new):
            self.__level = new

        @property
        def started(self):
            """
            Indicates whether the logging device has been started or not yet.

            Returns:
                * :bool:`True`:
                    Returned if the device has been started.

                * :bool:`False`:
                    Returned if the device has not yet been started.

            Note:
              :property:`InspyLogger().LogDevice().started` is a read-only property.

            """
            return self.__started



"""

File: inspy_logger/__about__.py
Project: inSPy-Logger
Description: 

Created: 11/3/22 - 21:05:06

Since:
    v2.1.1

"""

__prog__ = 'inSPy-Logger'
__version__ = '2.1.1'
__authors__ = [
        ('Inspyre-Softworks', 'https://inspyre.tech'),
        ('Taylor-Jayde Blackstone', '<t.blackstone@inspyre.tech')
]



"""
network.py

Description:
    Contains helper code pertaining to networks.

File:
    PROJECT_ROOT/inspy_logger/helpers/network.py

Project:
    Inspy-Logger

Author:
    Inspyre Softworks
        - Taylor-Jayde Blackstone <t.blackstone@inspyre.tech>

Since Version:
    2.1

Date:
    10/25/2022 - 03:28hrs

"""

# Do our import
from urllib.parse import quote as url_safe

import requests

from inspy_logger.helpers.cache import TLDCache

DEFAULT_TEST_HOSTS = [
        ['https', 'inspyre', 'tech'],
        ['https', 'google', 'com']
]
"""
:obj:`list`[:obj:`list`[:obj:`str`]]:
        The hosts provided by default to test connectivity against.
"""

hosts = []
"""
:obj:`list`[:obj:`Host`]: 
    A list of host objects that will be used to connect to. 
"""

TLD_CACHE = TLDCache()

VALID_TLDS = TLD_CACHE.TLDs


class Host():
    def __init__(self, host: list):

        # Populate the .__incoming variable with the URL parts
        # provided by the 'host' parameter
        self.__incoming = host

        # Set placeholders for all other host properties.
        self.__host_name = None
        self.__status = None
        self.__URL = None

    class URLObj():
        """
        A class to contain important URL properties.
        """

        def __init__(self, url_parts: list[str]):
            """
            Instantiate a new URLObj instance.



            Args:
                url_parts (:class:`list[:obj:``str``]`):
                    Pass in a list of url parts that will be unpacked into the class properties

            Returns:
                :class:`URLObj`

            """
            # Unpack our incoming URL parts into protected variables that
            # will be accessible via class properties.
            self.__protocol, self.__host_name, self.__tld = url_parts
            self.__valid_protocols = [
                    'http',
                    'https',
            ]

        @property
        def host_name(self) -> str:
            """ Return the host name associated with this instance.

            The value of this property is filled by the the contents of the 'url_parts'
            parameter (provided on instantiation) or via this property's setter.

            Returns:
                str:
                    The host_name associated with this instance.

            Raises:
                :exception:`TypeError` if the value of this property is not set to a string.

            """
            return self.__host_name

        @host_name.setter
        def host_name(self, new):
            if not isinstance(new, str):
                raise TypeError('The new value must be a string!')

            new = url_safe(new)

            self.__host_name = new

        @property
        def protocol(self) -> str:
            """
            str: The URLs protocol string.

            Note:
                This property must be set to a string value that is one of the following;

                    * 'http'
                    * 'https'

            Raises:
                :exception:`ValueError` if the value is not set to a valid protocol string.

            """
            return self.__protocol

        @protocol.setter
        def protocol(self, new: str):
            if new.lower() not in self.valid_protocols:
                raise ValueError(
                    f'The protocol must be one of {",".join(self.__valid_protocols)}.')

            self.__protocol = new

        @property
        def valid_protocols(self):
            """
            list: A list of valid protocol strings.
            """
            return self.__valid_protocols

        @property
        def valid_top_level_domains(self):
            """
            :obj:`list` of :obj:`str`:  A list of valid top-level domain strings.
            """
            return VALID_TLDS

        @property
        def tld(self) -> str:
            """
            The tld function returns the top-level domain of a given URL.

            Returns:
                str: The top level domain of the url

            """
            return self.__tld

        @tld.setter
        def tld(self, new):
            if not isinstance(new, str):
                raise TypeError(f"New value must be a string not {type(new)}")
            if new not in TLD_CACHE.TLDs:
                raise ValueError("TLD must be valid!")

            self.__tld = new

        @property
        def formatted(self):
            """
            The formatted function returns a string of the URL in the format:
            protocol://host_name.tld

            Example:
                Use this when you wanna get the URL in string format::
                    >>> from inspy_logger.helpers.network import Host
                    >>> host = Host(['https', 'inspyre', 'tech'])
                    >>> print(host.URL.formatted)

                    'https://inspyre.tech'

            Returns:
                A string that is the concatenation of all 3 URL parts;

                - Protocol
                - Host Name,
                - Top level domain (TLD)


            """
            return f"{self.protocol}://{self.host_name}.{self.tld}"

        @property
        def as_dict(self):
            """
            The as_dict function returns a dictionary containing the following:
                - formatted:
                    The URL as a string, e.g. 'http://google.com'
                - parts:
                    A dictionary containing the protocol, host_name and TLD of the URL
                        - protocol:
                            The protocol used to access this host (e.g., http)
                        - host_name:
                            The name of the server that hosts this host (e.g., google)
                        - TLD:
                            The top-level domain for this host (e.g., com)

            Args:
                self: Refer to the object itself

            Returns:
                A dictionary with the following keys:

                * formatted
                * parts
                    - protocol
                    - host_name
                    - TLD
                * reachable
            """
            return dict(
                {
                        'formatted': self.formatted,
                        'parts':     {
                                'protocol':  self.protocol,
                                'host_name': self.host_name,
                                'TLD':       self.tld
                        },
                        'reachable': None
                }
            )

    @property
    def host_name(self) -> (str | None):
        """ Return the name of the host. """

        if self.__host_name is None:
            self.__host_name = self.__incoming[1].lower()

        return self.__host_name

    @property
    def valid_status_codes(self) -> list:
        """ Returns the valid status codes for the host. """
        return ['UP', 'DOWN']

    @property
    def status(self) -> str:
        """
        The online status of the host.

        Can be one of:
            * UP:
                This would indicate that we've tested connectivity to this host and were able to establish a connection.

            * DOWN:
                This would indicate that we've tested the connectivity to this host and were unable to establish a connection.

        """
        return self.__status

    @status.setter
    def status(self, new):
        if new.upper() not in self.valid_status_codes:
            raise ValueError('Invalid status code')

        self.__status = new.upper()

    @property
    def can_connect(self) -> (bool | None):
        """
        The can_connect function is a helper function that checks if the server is up and running.
        It returns True if it can connect to the host, False otherwise.


        Returns:
            A boolean value

        """
        if self.status is not None:
            return self.status != 'DOWN'

    @property
    def URL(self):
        """
        The URL property returns the URL property.

        Returns:
            :class:`inspy_logger.helpers.networks.Host.URLObj`

        """
        if self.__URL is None:
            self.__URL = self.URLObj(self.__incoming)

        return self.__URL

    def check_connectivity(self, timeout):
        """
        The check_connectivity function checks the connectivity of a given URL.
        It returns 'UP' if the connection is successful, and 'DOWN' otherwise.

        Parameters:
            timeout (int):
                The number of seconds to wait for the connection to be established. (Required)

        Returns:
            The status of the host:

                - "UP":
                    The host is reachable.

                - "DOWN":
                    The host is unreachable.
        """
        try:
            requests.head(
                self.URL.formatted,
                timeout = timeout
            )
            self.status = 'UP'

        except requests.ConnectionError:
            self.status = 'DOWN'

        return self.status


def check_connectivity(
        test_hosts=None,
        timeout=3,
        online_if_any=True,
        online_if_all=False
):
    """
    The check_connectivity function checks the connectivity of a list of hosts.

    Parameters:
        test_hosts (list):
            A list of hosts to test connectivity to.

        timeout (int):
            The number of seconds to wait for a connection to be established. (Defaults to 3)

        online_if_any (bool):
            Whether the function should consider us online if any of the hosts are connectable.
            (Defaults to True)

        online_if_all (bool):
            Whether the function should only consider us online if all of the hosts are connectable.
            (Defaults to False)

    Returns:
        bool:
            True: Connectivity check passed successfully.
            False: Connectivity check failed.
    """
    if test_hosts is None:
        test_hosts = DEFAULT_TEST_HOSTS

    for host in test_hosts:
        hosts.append(Host(host))

    if online_if_any and online_if_all:
        raise ValueError('Both online_if_any and online_if_all cannot be True')

    statuses = [host.check_connectivity(timeout = timeout) for host in hosts]

    if online_if_any:
        return 'UP' in statuses

    if online_if_all:
        return 'DOWN' in statuses



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inspy_logger', 'inspy_logger.errors', 'inspy_logger.helpers']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'colorama>=0.4.6,<0.5.0',
 'colorlog>=6.7,<7.0',
 'domain-suffixes>=1.1.3,<2.0.0',
 'humanize>=4.4.0,<5.0.0',
 'luddite>=1.0.1,<2.0.0',
 'packaging>=21.0,<22.0',
 'python-box>=6.1.0,<7.0.0',
 'requests>=2.28.1,<3.0.0',
 'setuptools-autover>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'inspy-logger',
    'version': '2.1.1',
    'description': 'Colorable, scalable logger for CLI',
    'long_description': None,
    'author': 'Taylor-Jayde Blackstone',
    'author_email': 'tayjaybabee@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from time import time
from logging import getLogger
from box import Box


class Manifest(object):
    def __init__(self, root_name, root_device, calling_file, line_num):
        self.name = root_name

        self.contents = None
        self.add(root_name, root_device, calling_file, line_num)

    def check(self, name):

        return self.contents[name]['device'] if name in self.contents.keys() else None


    def add(self, name, logger_device, calling_file, line_num):
        """

        Add the provided logger to the manifest.

        Parameters:
          name (str): The logger name.
          logger_device ()

        """

        entry = {
                name: {
                        'created-ts':  time(),
                        'device':      logger_device,
                        'caller_file': calling_file,
                }
        }

        if self.contents is None:
            self.contents = Box(entry)
        else:
            self.contents.update(entry)
